/* ----------------------------------------------------------------------------
 *     Copyright (c) 2000-2019, Nexus Electronics Ltd (www.nexus.co.uk)
 *                            All rights reserved
 *
 * Redistribution and use in source and derived forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in derived forms must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of Nexus Electronics Ltd. nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY NEXUS ELECTRONICS LTD. "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NEXUS ELECTRONICS LTD. BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * ----------------------------------------------------------------------------
 * Title       : 3-DES, controller module
 * Project     : MAS
 * RCS ID      : 
 * $Id: controller.v,v 1.2 2001/01/16 18:13:52 el Exp $
 * ----------------------------------------------------------------------------
 * File        : controller.v
 * Author      : E.M. Lavelle
 * Company     : Nexus Electronics Ltd.
 * Last saved  : 22/08/2000
 * Version     : $Revision: 867 $
 * ----------------------------------------------------------------------------
 * Description : 
 * 
 * this module carries out all control ops for triple-DES. it generates 16
 * rounds by providing a clock enable, and repeats the rounds 3 times, in
 * either an encode-decode-encode or a decode-encode-decode sequence,
 * together with the appropriate key selection, for 2- or 3-key triple-DES.
 * 
 * ----------------------------------------------------------------------------
 * Modification history:
 * Date        Version Author     Comment
 * 22/08/2000  0.1      EML       created
 *
 * ------------------------------------------------------------------------- */

module controller
   // in
  (CLK,                         // clock
   SRST_H,                      // sync reset
   START_H,                     // data register load
   ENC3D_H,                     // encrypt/decrypt mode
   // out
   KSEL,			// select key
   KLD_H,			// key load strobe
   DLD_H,	                // data load strobe 
   DCE_H,			// data clock enable
   SWAP_H,			// swap left/right
   ENC_H,			// current round encrypt/decrypt
   DRDY1_H,			// data ready
   DRDY2_H			// DRDY1_H delayed 1 cycle
   );

   input CLK, SRST_H, START_H, ENC3D_H;
   output [1:0]  KSEL;
   reg [1:0] 	 KSEL;
   output KLD_H, DLD_H, DCE_H, SWAP_H, ENC_H, DRDY1_H, DRDY2_H;
   reg 	  KLD_H, ENC_H, DRDY2_H;

   // wait controller counter
   reg 	  WAIT_C0_H;

   // wait controller
   wire WAIT_START_H;
   wire WAIT_DCE_H, WAIT_TERM_H, WAIT_CLR_H, WAIT_INCR_H;
   reg [3:0] WAIT_SREG;
   parameter WS0  = 4'h0, WS1  = 4'h8, WS2  = 4'h9, WS3  = 4'h1,
	     WS4  = 4'hA, WS5  = 4'h2, WS6  = 4'hB, WS7  = 4'h3,
	     WS8  = 4'hC, WS9  = 4'h4, WS10 = 4'hD, WS11 = 4'h5,
	     WS12 = 4'hE, WS13 = 4'h6, WS14 = 4'h7;

   // 3-DES controller
   wire KSEL_C3_H;		             // inputs
   wire KSEL_CLR_H, KSEL_INCR_H, DES_DCE_H;  // outputs
   reg [6:0] DES_SREG;
   parameter DS0 = 7'h40, DS1 = 7'h20, DS2 = 7'h0c, DS3 = 7'h00,
	     DS4 = 7'h14, DS5 = 7'h04, DS6 = 7'h06, DS7 = 7'h01;

   // combinatorial outputs
   assign DCE_H = WAIT_DCE_H | DES_DCE_H;

   // delay DRDY1_H one cycle as DRDY2_H. DRDY1_H loads the output data
   // into R3, and DRDY2_H tells the rest of the world that the output
   // is ready
   always @(posedge CLK)
     if(SRST_H)
          DRDY2_H <= 0;
     else DRDY2_H <= DRDY1_H;

   // generate the key load strobe
   always @(posedge CLK)
     begin
	if(SRST_H)			// sync reset
	  KLD_H <= 0;
	else
	  KLD_H <= START_H |
		  (~ENC3D_H & KSEL_CLR_H) |
		  (~ENC3D_H & (KSEL == 2) & (DES_SREG == DS4)) |
		  (~ENC3D_H & (KSEL == 0) & (DES_SREG == DS5)) |
		  ( ENC3D_H & (KSEL == 1) & (DES_SREG == DS5)) |
		  ( ENC3D_H & (KSEL == 1) & (DES_SREG == DS4));
     end
       
   // generate the EDE or DED sequence for the key module
   always @(posedge CLK)
     if(SRST_H)			// sync reset
       ENC_H <= 0;
     else if(KSEL_CLR_H)	// load
       ENC_H <= ENC3D_H;
     else if(KSEL_INCR_H)	// toggle
       ENC_H <= ~ENC_H;
   
   /* ----------------------------------------------------------------------
    * 3-DES CONTROLLER
    * ---------------------------------------------------------------------- */

   // 3-DES controller registered outputs
   assign KSEL_CLR_H  = DES_SREG[5];      // KSEL counter clear
   assign KSEL_INCR_H = DES_SREG[4];      // KSEL counter increment
   assign DLD_H       = DES_SREG[3];      // module output
   assign DES_DCE_H   = DES_SREG[2];
   assign SWAP_H      = DES_SREG[1];      // module output
   assign DRDY1_H     = DES_SREG[0];      // module output

   // KSEL counter. this selects the required key for the current phase
   // of a 3DES op. basically, this preloads either 0 or 2, and then
   // increments for encodes, and decrements for decodes. the complication
   // with START_H is a timing issue: the preload must occur immediately
   // for encoding, but can be corrected one cycle later for decoding.
   // START_H
   //  KSEL_CLR_H
   //   KSEL_INCR_H
   //    ENC3D_H
   // 1xxx  preload 0
   // 01x0  decode, preload 2
   // 01x1  hold
   // 0010  decode, decrement
   // 0011  encode, increment
   // 000x  hold
   always @(posedge CLK)
     begin
	if(SRST_H | START_H)
	  KSEL <= 0;		// sync reset, and preload for increment
	else if(KSEL_CLR_H & ~ENC3D_H)
	  KSEL <= 2;		// corrected preload for decrement
	else if(KSEL_INCR_H & ENC3D_H)
	  KSEL <= KSEL + 1;	// increment through the 3 keys
	else if(KSEL_INCR_H & ~ENC3D_H)
	  KSEL <= KSEL - 1;	// decrement through the 3 keys
     end

   // KSEL terminal count: C3 is decoded for both increment and decrement
   assign KSEL_C3_H = (KSEL == 3);

   // 3-DES controller
   always @(posedge CLK)
     begin
	if(SRST_H)
	  DES_SREG <= DS0;
	else
	  case(DES_SREG)
	    DS0 :
	      if(START_H)
		DES_SREG <= DS1;
	    DS1 : DES_SREG <= DS2;
	    DS2 : DES_SREG <= DS3;
	    DS3 :
	      if(WAIT_TERM_H)
		DES_SREG <= DS4;
	    DS4 : DES_SREG <= DS5;
	    DS5 : DES_SREG <= DS6;
	    DS6 :
	      if(KSEL_C3_H)
		   DES_SREG <= DS7;
	      else DES_SREG <= DS3;
	    default : DES_SREG <= DS0;        // DS7
	  endcase // case(DES_SREG)
     end
   
   /* ----------------------------------------------------------------------
    * WAIT CONTROLLER
    * ---------------------------------------------------------------------- */

   // 1-bit counter for the wait controller
   always @(posedge CLK)
     begin
	if(SRST_H | WAIT_CLR_H)
	  WAIT_C0_H <= 0;
	else if(WAIT_INCR_H)
	  WAIT_C0_H <= 1;
     end

   // wait controller start condition
   assign WAIT_START_H = DLD_H | (SWAP_H & (KSEL != 3));
   
   // wait controller registered outputs
   assign WAIT_DCE_H  = WAIT_SREG[3];

   // wait controller combinatorial outputs
   assign WAIT_CLR_H  = (WAIT_SREG == WS0)  & WAIT_START_H;
   assign WAIT_INCR_H = (WAIT_SREG == WS12) & ~WAIT_C0_H;
   assign WAIT_TERM_H = (WAIT_SREG == WS13);

   // the wait controller
   always @(posedge CLK)
     begin
	if(SRST_H)
	  WAIT_SREG <= WS0;
	else
	  case(WAIT_SREG)
	    WS0 : 
	      if(WAIT_START_H)
		WAIT_SREG <= WS1;
	    WS1  : WAIT_SREG <= WS2;
	    WS2  : WAIT_SREG <= WS3;
	    WS3  : WAIT_SREG <= WS4;
	    WS4  : WAIT_SREG <= WS5;
	    WS5  : WAIT_SREG <= WS6;
	    WS6  : WAIT_SREG <= WS7;
	    WS7  : WAIT_SREG <= WS8;
	    WS8  : WAIT_SREG <= WS9;
	    WS9  : WAIT_SREG <= WS10;
	    WS10 : WAIT_SREG <= WS11;
	    WS11 : WAIT_SREG <= WS12;
	    WS12 : 
	      if(WAIT_C0_H)
		   WAIT_SREG <= WS13;
	      else WAIT_SREG <= WS14;
	    WS14 : WAIT_SREG <= WS1;
	    default : WAIT_SREG <= WS0;         // WS13
	  endcase // case(WAIT_SREG)
     end
endmodule  

/*----------------------------------- EOF -----------------------------------*/
